/* hammerhead.c

	creates a /proc entry to get details
	on the hammerhead controller
	and the physical memory configuration

	quick&dirty hack by M. Lanners <mlan@cpu.lu>
	
	based on code by Apple Computer, Inc.
*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/prom.h>

#define interleaveMask 		0x04
#define adrs1Mask               0x01
#define adrs1Shift              8 
#define MegPerAdrs              4

int hammerhead_read_procmem(char *, char **, off_t, int, int);

int	RAMSize=0;	/* Parameter!" */
int	Model=0;	/* Parameter!" */

MODULE_PARM(RAMSize, "i");
MODULE_PARM(Model, "i");

struct proc_dir_entry hammerhead_proc_entry = {
	0,
	10, "hammerhead",
	S_IFREG | S_IRUGO,
	1, 0, 0,
	0,
	NULL,
	&hammerhead_read_procmem,
};

struct RAMBankDescriptor {
        unsigned long		bankSize;
        int			bankIsInterleaved;
};

typedef struct RAMBankDescriptor RAMBankDescriptor;
typedef RAMBankDescriptor* RAMBankDescriptorPtr;

void GetPhysicalRAMBankInfo(RAMBankDescriptorPtr  theBanks, unsigned long  fullMemSize, unsigned long hammerHeadBase)
{
        char *		bankN;
        short		i;
        
        bankN = (char *)hammerHeadBase + 0x1C0;
    
        for (i=0; i<=25; i++)
        {
                theBanks[i].bankIsInterleaved = (*bankN) & interleaveMask;
                theBanks[i].bankSize = (*bankN) & adrs1Mask ;
                bankN = bankN + 0x10;
                theBanks[i].bankSize = (theBanks[i].bankSize << adrs1Shift) + (*
bankN);
                bankN = bankN + 0x10;
        } 
        
        for (i = 0; i < 25; i++)
        {
                theBanks[i].bankSize = (theBanks[i+1].bankSize - theBanks[i].bankSize) * MegPerAdrs;
        }

        theBanks[25].bankSize = (fullMemSize / 4 - theBanks[25].bankSize) * MegPerAdrs;

        for (i = 0; i < 25; i++)
        {
                if (theBanks[i].bankIsInterleaved) {
                        theBanks[i+1].bankSize = theBanks[i].bankSize / 2;
                        theBanks[i].bankSize = theBanks[i+1].bankSize;
                }
        }       
        
        return;
}

static void
PrintBankInfo(int k, RAMBankDescriptorPtr theBanks, char *buf, int *len)
{
        (*len)+= sprintf (buf+(*len), "%3.0d/%3.0d ", theBanks[k].bankSize, theBanks[k+1].bankSize);

        if (theBanks[k].bankIsInterleaved)
                (*len)+= sprintf (buf+(*len), "Interleaved");
        else
                (*len)+= sprintf (buf+(*len), " --- ");

        (*len)+= sprintf(buf+(*len), "\n");
}

static void
Print75008500 (RAMBankDescriptorPtr  theBanks, char *buf, int *len)
{
	int	i;
	int	j;

	(*len)+= sprintf (buf+(*len), "Motherboard Row-> not implemented\n"); 

        for (i=1; i<=4; i++)    
        {
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 0) -> ", i, i);   
                j = 18-(4*i);
                PrintBankInfo(j, theBanks, buf, len);
        
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 1) -> ", i, i);   
                j = 20-(4*i);
                PrintBankInfo(j, theBanks, buf, len);
        }
}

static void 
Print9500 (RAMBankDescriptorPtr  theBanks, char *buf, int *len)
{
	int	i, j;

	(*len)+= sprintf (buf+(*len), "Motherboard Row-> not implemented\n"); 

        // Decipher the bank information, and display it.
        for (i=1; i<=6; i++)            
        {
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 0) -> ", i, i);   
                j = (4*i)-2;
                PrintBankInfo(j, theBanks, buf, len);
                
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 1) -> ", i, i);   
                j = 4*i;
                PrintBankInfo(j, theBanks, buf, len);
        }
}               

static void
PrintGenericMac (RAMBankDescriptorPtr  theBanks, char *buf, int *len)
{
	int	i, j;

	for (i = 0; i < 13; i++)
	{
		(*len)+= sprintf (buf+(*len), "x%d/x%d         -> ", i, i);
		j = 2 * i;
		PrintBankInfo(j, theBanks, buf, len);
	}
}

unsigned long findhammerheadbase(void)
{
	struct  device_node *dev;
	unsigned long 	addr;
	
	dev = find_devices("hammerhead");
	if (dev == 0)
		return 0;
	
	addr = ioremap (dev->addrs[0].address, 0x400);
	return addr;
}

int hammerhead_read_procmem(char *buf, char **start, off_t offset, int len, int unused)
{
	unsigned long	hammerheadbase;
	extern int	RAMSize;
	extern int	Model;
	RAMBankDescriptor	allBanks[26];

	len=0;

	if ((hammerheadbase = findhammerheadbase()) == 0) {
		len += sprintf (buf+len, "---Unsupported Macintosh!---\n");
		return len;
	}

	if (RAMSize == 0) {
		len += sprintf (buf+len,"---Memory Size unspecified!---\n");
		return len;
	}
	else
		len += sprintf (buf+len,"Total Memory = %d Megs in \n", RAMSize);
	GetPhysicalRAMBankInfo(allBanks, RAMSize, hammerheadbase);

	switch (Model)
	{
	  case 0:
		len += sprintf (buf+len,"---Model unspecified!---\n");
		return len;
		break;
	  case 7300:
		len += sprintf (buf+len,"7300 DIMM Slots: Meg/Meg Interleaved?\n");
		Print75008500(allBanks, buf, &len);
		break;
	  case 7500:
		len += sprintf (buf+len,"7500 DIMM Slots: Meg/Meg Interleaved?\n");
		Print75008500(allBanks, buf, &len);
		break;
	  case 7600:
		len += sprintf (buf+len,"7600 DIMM Slots: Meg/Meg Interleaved?\n");
		Print75008500(allBanks, buf, &len);
		break;
	  case 8500:
		len += sprintf (buf+len,"8500 DIMM Slots: Meg/Meg Interleaved?\n");
		Print75008500(allBanks, buf, &len);
		break;
	  case 8600:
		len += sprintf (buf+len,"8600 DIMM Slots: Meg/Meg Interleaved?\n");
		Print75008500(allBanks, buf, &len);
		break;
	  case 9500:
		len += sprintf (buf+len,"9500 DIMM Slots: Meg/Meg Interleaved?\n");
		Print9500(allBanks, buf, &len);
		break;
	  case 9600:
		len += sprintf (buf+len,"9600 DIMM Slots: Meg/Meg Interleaved?\n");
		Print9500(allBanks, buf, &len);
		break;
	  default:
		len += sprintf (buf+len,"Unknown PowerMac: Meg/Meg Interleaved?\n");
		PrintGenericMac(allBanks, buf, &len);
		break;
	}

	return len;
}

#ifdef MODULE

int init_module(void)
{
	/* other stuff */

	proc_register(&proc_root, &hammerhead_proc_entry);

	return 0;
}

void cleanup_module(void)
{
	proc_unregister(&proc_root, hammerhead_proc_entry.low_ino);
}

#endif /* MODULE */
